// $Id: TCP_Rewriter.cc,v 1.18 2002/10/26 04:15:44 vern Exp $

//  Overview of TCP trace rewriter:
//
//  1. Timestamp: Consider every packet arrival at a certain endpoint
//  in the original trace as a tick for the endpoint; a packet will be
//  dumped into the new trace (possibly with different content) in the
//  tick (enforced by flush_rewriter_packet). When the user writes data
//  into the new trace, the packet will carry a timestamp of the
//  current tick, or the next tick if data is generated between
//  ticks. Users may also choose to Push data, in which case the packet
//  carries a timestamp of current network time. This gives us the
//  'freshness' of timestamps.
//
//  2. Ordering of contents: user may choose to push contents in order
//  to enforce ordering of contents between two directions, however,
//  contents MIGHT be already dumped BEFORE they are pushed, once they
//  are written into the trace. (similar to PUSH in TCP)
//
//  3. Acknowledgements: when a packet is dumped at a time when there
//  is no corresponding packet in the original trace, an articial
//  acknowledgement is generated from the peer with the same
//  timestamp. This guarantees that additional packets will have
//  acknowledgements. For those packets dumped at 'ticks', there
//  should be corresponding acknowledgement packets in the original
//  trace already, so we do not generate further artificial
//  acknowledgement packets.
//
//  4. SYN, RST, FIN: SYN/RST packets in the new trace do not carry
//  payloads -- additional packets may be generated for payloads. FIN
//  packets are generated only when user calls ScheduleFIN, which by
//  default corresponds to be the moment a FIN appears in the original
//  trace. Change: now we will try to allow SYN/RST packets to carry
//  payloads if they originally do.

#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "Event.h"
#include "Net.h"
#include "TCP_Rewriter.h"

#define MSG_PREFIX	"TCP trace rewriter: "

static struct timeval double_to_timeval(double t)
	{
	struct timeval tv;

	double t1 = floor(t);
	tv.tv_sec = (int) t1;

	// Round to integer to tolerate floating arithmetic
	// imprecision (which can be up to 0.05 here!).
	tv.tv_usec = (int) ((t - t1) * 1e6 + 0.5);

	return tv;
	}

inline int time_compare(struct timeval* tv_a, struct timeval* tv_b)
	{
	if ( tv_a->tv_sec == tv_b->tv_sec )
		return tv_a->tv_usec - tv_b->tv_usec;
	else
		return tv_a->tv_sec - tv_b->tv_sec;
	}

PacketDumper::PacketDumper(pcap_dumper_t* arg_pkt_dump)
	{
	last_ts.tv_sec = last_ts.tv_usec = 0;
	pkt_dump = arg_pkt_dump;
	if ( ! pkt_dump )
		internal_error(MSG_PREFIX "NULL dump file");
	}

PacketDumper::~PacketDumper()
	{
	if ( pkt_dump )
		pcap_dump_close(pkt_dump);
	}

void PacketDumper::DumpPacket(const struct pcap_pkthdr* hdr,
				const u_char* pkt, int len)
	{
	if ( pkt_dump )
		{
		struct pcap_pkthdr h = *hdr;
		h.caplen = len;
		if ( h.caplen > hdr->caplen )
			internal_error("bad modified caplen");

#if 0
// Do not do the check any more because packets may need be
// delayed for commitment.
		// For debug only: catch packets with timestamp not
		// equal to network_time.
		struct timeval now = double_to_timeval(network_time);
		if ( now.tv_sec != h.ts.tv_sec || now.tv_usec != h.ts.tv_usec)
			{
			DEBUG_MSG("%ld.%06ld != %ld.%06ld\n",
				  now.tv_sec, now.tv_usec,
				  h.ts.tv_sec, h.ts.tv_usec);
			internal_error(MSG_PREFIX "timestamp inconsistency");
			}
#endif

		pcap_dump((u_char*) pkt_dump, &h, pkt);
		}
	}

void PacketDumper::SortTimeStamp(struct timeval* ts)
	{
	if ( time_compare(&last_ts, ts) > 0 )
		*ts = last_ts;
	else
		last_ts = *ts;
	}


TCP_TracePacket::TCP_TracePacket(double t, const struct pcap_pkthdr* arg_hdr, int MTU)
	{
	pcap_hdr = *arg_hdr;
	ts = t;
	pkt = new u_char[MTU];
	buffer_size = MTU;

	buffer_offset = 0;
	ip_offset = tcp_offset = data_offset = -1;
	reuse = 0;
	}

TCP_TracePacket::~TCP_TracePacket()
	{
	delete [] pkt;
	}

int TCP_TracePacket::AppendLinkHeader(const u_char* chunk, int len)
	{
	if ( Space() < len )
		return 0;

	if ( ip_offset >= 0 && ip_offset != buffer_offset )
		internal_error(MSG_PREFIX "link header must be appended before ip header");

	Append(chunk, len);
	ip_offset = buffer_offset;

	return 1;
	}

int TCP_TracePacket::AppendIPHeader(const u_char* chunk, int len)
	{
	if ( Space() < len )
		return 0;

	if ( tcp_offset >= 0 && tcp_offset != buffer_offset )
		internal_error(MSG_PREFIX "ip header must be appended before tcp header");

	Append(chunk, len);
	tcp_offset = buffer_offset;

	return 1;
	}

int TCP_TracePacket::AppendTCPHeader(const u_char* chunk, int len)
	{
	if ( Space() < len )
		return 0;

	if ( data_offset >= 0 && data_offset != buffer_offset )
		internal_error(MSG_PREFIX "tcp header must be appended before payload");

	if ( tcp_offset == buffer_offset )
		{ // first TCP header chunk
		int extra = (tcp_offset - ip_offset) % 4;
		if ( extra )
			{
			DEBUG_MSG(MSG_PREFIX "padding ip header");
			AppendIPHeader(0, 4 - extra);
			}
		}

	Append(chunk, len);
	data_offset = buffer_offset;

	return 1;
	}

int TCP_TracePacket::AppendData(const u_char* chunk, int len)
	{
	if ( Space() < len )
		return 0;

	// All headers must be appended before any data.
	ASSERT(ip_offset >= 0 && tcp_offset >= 0 && data_offset >= 0);

	if ( data_offset == buffer_offset )
		{ // first data chunk
		int extra = (data_offset - tcp_offset) % 4;
		if ( extra )
			{
			DEBUG_MSG(MSG_PREFIX "padding tcp header");
			AppendTCPHeader(0, 4 - extra);
			}
		}

	Append(chunk, len);

	return 1;
	}

void TCP_TracePacket::Append(const u_char* chunk, int len)
	{
	ASSERT(buffer_offset + len <= buffer_size);

	if ( chunk )
		memcpy(pkt + buffer_offset, chunk, len);
	else
		// fill with 0
		memset(pkt + buffer_offset, 0, len);

	buffer_offset += len;
	}

uint32 TCP_TracePacket::GetSeq() const
	{
	ASSERT(tcp_offset >= ip_offset + (int)sizeof(struct ip) &&
	       buffer_offset >= tcp_offset + (int)sizeof(struct tcphdr));

	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);
	return ntohl(tp->th_seq);
	}

void TCP_TracePacket::SetSeq(uint32 seq)
	{
	ASSERT(tcp_offset >= ip_offset + (int)sizeof(struct ip) &&
	       buffer_offset >= tcp_offset + (int)sizeof(struct tcphdr));

	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);
	tp->th_seq = htonl(seq);
	}

uint32 TCP_TracePacket::GetAck() const
	{
	ASSERT(tcp_offset >= ip_offset + (int)sizeof(struct ip) &&
	       buffer_offset >= tcp_offset + (int)sizeof(struct tcphdr));

	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);
	return ntohl(tp->th_ack);
	}

void TCP_TracePacket::SetAck(uint32 ack)
	{
	ASSERT(tcp_offset >= ip_offset + (int)sizeof(struct ip) &&
	       buffer_offset >= tcp_offset + (int)sizeof(struct tcphdr));

	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);
	tp->th_ack = htonl(ack);
	}

int TCP_TracePacket::GetTCPFlag(int which) const
	{
	ASSERT(tcp_offset >= ip_offset + (int)sizeof(struct ip) &&
	       buffer_offset >= tcp_offset + (int)sizeof(struct tcphdr));

	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);
	return tp->th_flags & which;
	}

void TCP_TracePacket::SetTCPFlag(int which, int value)
	{
	ASSERT(tcp_offset >= ip_offset + (int)sizeof(struct ip) &&
	       buffer_offset >= tcp_offset + (int)sizeof(struct tcphdr));

	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);

	if ( value )
		tp->th_flags |= which;
	else
		tp->th_flags &= (~which);
	}

int TCP_TracePacket::PayloadLength() const
	{
	if ( data_offset < 0 )
		return 0;

	return buffer_offset - data_offset;
	}

int TCP_TracePacket::SeqLength() const
	{
	int len = PayloadLength();
	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);

	if ( tp->th_flags & TH_SYN )
		++len;

	if ( tp->th_flags & TH_FIN )
		++len;

	return len;
	}

int TCP_TracePacket::Finish(struct pcap_pkthdr*& hdr,
			    const u_char*& arg_pkt, int& length,
			    ipaddr32_t anon_src, ipaddr32_t anon_dst)
	{
	// Set length fields in headers and compute checksums.
	if ( tcp_offset < ip_offset + (int)sizeof(struct ip) ||
	     data_offset < tcp_offset + (int)sizeof(struct tcphdr) )
		return 0;

	struct ip* ip = (struct ip*) (pkt + ip_offset);
	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);

	// TCP header.
	ASSERT((data_offset - tcp_offset) % 4 == 0);

	tp->th_off = (data_offset - tcp_offset) >> 2;
	tp->th_x2 = 0;
	tp->th_flags &= (~TH_URG);	// set URG to 0
	tp->th_urp = 0;			// clear urgent pointer
	tp->th_sum = 0;
	tp->th_sum = 0xffff - tcp_checksum(ip, tp, PayloadLength());

	// IP header.
	if ( anonymize_ip_addr )
		{
		ip->ip_src.s_addr = anon_src;
		ip->ip_dst.s_addr = anon_dst;
		}

	ASSERT((tcp_offset - ip_offset) % 4 == 0);
	ip->ip_hl = (tcp_offset - ip_offset) >> 2;
	ip->ip_len = htons(buffer_offset - ip_offset);
	ip->ip_off = 0;		// DF = 0, MF = 0, offset = 0
	ip->ip_sum = 0;
	ip->ip_sum = 0xffff - ones_complement_checksum((const void*)ip, tcp_offset - ip_offset, 0);

	// Link level header:
	// Question: what to do link level header? Currently we just
	// keep the original header, even though the length field can
	// be incorrect. ###

	// PCAP header
	ts = network_time;

	pcap_hdr.ts = double_to_timeval(ts);
	pcap_hdr.caplen = pcap_hdr.len = buffer_offset;

	hdr = &pcap_hdr;
	arg_pkt = pkt;
	length = buffer_offset;

	return 1;
	}

void TCP_TracePacket::Reuse()
	{
	reuse = 1;

	ts = network_time;

	// Clear data in the packet, and headers can be reused.
	buffer_offset = data_offset;

	// Question 1: Shall we keep TCP options in the header? Note
	// that the TCP header of a packet is sometimes replicated in
	// the rewritten trace (because we reuse headers). When an
	// option have idempotent semantics, it is safe to keep the
	// option; otherwise we should include the option only in the
	// first one among replicated copies. Currently we keep
	// options for simplicity and wait for things to happen. ###

	struct tcphdr* tp = (struct tcphdr*) (pkt + tcp_offset);
	tp->th_flags = 0;

	// Question 2: What to do with ip_id? Currently we choose a
	// pseudo-random number as the new id.
	struct ip* ip = (struct ip*) (pkt + ip_offset);
	ip->ip_id = (random() & 0xffff);
	}


TCP_RewriterEndpoint::TCP_RewriterEndpoint(TCP_Rewriter* arg_rewriter)
	{
	rewriter = arg_rewriter;
	next_packet = 0;
	endp = 0;
	established = 0;
	end_of_data = 0;
	peer = 0;
	last_ack = 0;
	last_packet_time = -1;
	please_flush = 0;
	flush_scheduled = 0;
	FIN_scheduled = 0;
	there_is_a_gap = 0;
	}

TCP_RewriterEndpoint::~TCP_RewriterEndpoint()
	{
	if ( ! end_of_data && next_packet && ! next_packet->Empty() )
		Weird(fmt(MSG_PREFIX "end of data missing before deleting the connection: %.6f",
			  next_packet->TS()));

	if ( next_packet )
		delete next_packet;
	}

void TCP_RewriterEndpoint::Init()
	{
	// This cannot be put into constructor because it requires
	// existence of the peer.
	peer = rewriter->GetPeer(this);
	}

// NextPacket sets 'ticks' of packet dumping according to packet
// arrival in the original sequence.
void TCP_RewriterEndpoint::NextPacket(TCP_TracePacket* p)
	{
	please_flush = 1;
	last_packet_time = p->TS();

	// ### Ruoming, why do you need to special-case SYN/RST's?  For
	// historical reasons. :( Let me see what happens if we do not
	// push SYN/RST packets.  08/03: So far it seems to be fine
	// after I made the change. Let's observe a while further and
	// remove this comment if everything is fine.

	if ( ! endp )
		endp = rewriter->GetEndpoint(this);

	if ( endp->state == TCP_ESTABLISHED )
		established = 1;

	if ( ! next_packet || p->GetTCPFlag(TH_SYN) )
		{
		if ( ! p->GetTCPFlag(TH_SYN | TH_RST) &&
		     ! rewriter->Conn()->IsPartial() )
			Weird(MSG_PREFIX "first packet is not SYN or RST");

		start_seq = next_seq = p->GetSeq();
		}

	SetNextPacket(p);

	if ( ! please_flush )
		internal_error(MSG_PREFIX "why not flush?");

	ScheduleFlush();
	}

void TCP_RewriterEndpoint::WriteData(int len, const u_char* data)
	{
	if ( end_of_data & (END_BY_FIN | END_BY_RST) )
		{
		Weird(MSG_PREFIX "write after end of data");
		// abort();
		}

	if ( ! next_packet )
		{
		DEBUG_MSG("pushing %d bytes into prolog\n", len);
		prolog.push(new BroString(data, len, 0));
		}
	else
		{
		// Originally we did not send data along with SYN or RST.
		// if ( next_packet->GetTCPFlag(TH_SYN | TH_RST) )
		//	internal_error("SYN/RST packet not immediately flushed");

		// Question: shall we send data along with the ACK in the
		// connection's three way handshake? Here it may do so.

		DoWriteData(len, data);

		if ( please_flush )
			ScheduleFlush();
		}
	}

void TCP_RewriterEndpoint::SkipGap(int len)
	{
	next_seq += len;
	there_is_a_gap = 1;
	}

void TCP_RewriterEndpoint::Push()
	{
	if ( ! next_packet )
		return;

	next_packet->SetTCPFlag(TH_PUSH, 1);
	PushPacket();
	}

void TCP_RewriterEndpoint::ReqAck()
	{
	if ( ! next_packet )
		return;

	PushPacket();
	}

void TCP_RewriterEndpoint::Flush()
	{
	if ( ! next_packet )
		// This may happen after the code change in
		// TCP_Connection::NextPacket -- not every packet
		// reaches TCP rewriter
		// internal_error(MSG_PREFIX "flush before packet arrival");
		return;

	if ( FIN_scheduled )
		{
		GenerateFIN();
		EndOfData();
		FIN_scheduled = 0;
		}

	if ( please_flush )
		PushPacket();

	flush_scheduled = 0;
	}

void TCP_RewriterEndpoint::ScheduleFlush()
	{
	if ( ! flush_scheduled )
		{
		schedule_flush(this);
		flush_scheduled = 1;
		}
	}

void TCP_RewriterEndpoint::GenerateFIN()
	{
	if ( end_of_data & END_BY_FIN )
		return;

	end_of_data |= END_BY_FIN;
	// DEBUG_MSG("FIN at %.6f\n", network_time);

	if ( ! next_packet )
		Weird(MSG_PREFIX "FIN before packet arrival");
	else
		next_packet->SetTCPFlag(TH_FIN, 1);
	}

void TCP_RewriterEndpoint::EndOfData()
	{
	if ( ! next_packet )
		{
		if ( ! prolog.empty() )
			{
			Weird(MSG_PREFIX "end point has data, but hasn't got any packet.");
			while ( ! prolog.empty() )
				{
				delete prolog.front();
				prolog.pop();
				}
			}

		return;
		}

	if ( ! next_packet->Empty() )
		PushPacket();
	}

// "Please call GenerateFIN in Flush."
void TCP_RewriterEndpoint::ScheduleFIN()
	{
	FIN_scheduled = 1;
	ScheduleFlush();
	}

void TCP_RewriterEndpoint::Reset(int self)
	{
	if ( self )
		end_of_data |= END_BY_RST;
	else
		end_of_data |= END_BY_PEER_RST;
	}

void TCP_RewriterEndpoint::SetNextPacket(TCP_TracePacket* p)
	{
	if ( next_packet )
		{
		if ( ! next_packet->Empty() )
			internal_error(MSG_PREFIX "the packet should have been flushed");
			// PushPacket();

		delete next_packet;
		}

	next_packet = p;

	// Overwrite PUSH & FIN.
	next_packet->SetTCPFlag(TH_PUSH, 0);
	next_packet->SetTCPFlag(TH_FIN, 0);

	if ( ! prolog.empty() )
		{
		// Do not put prolog into SYN/RST packets
		if ( next_packet->GetTCPFlag(TH_SYN | TH_RST) )
			PushPacket();
		PurgeProlog();
		}
	}

void TCP_RewriterEndpoint::PurgeProlog()
	{
	while ( ! prolog.empty() )
		{
		BroString* s = prolog.front();
		WriteData(s->Len(), s->Bytes());
		prolog.pop();
		delete s;
		}
	}

void TCP_RewriterEndpoint::DoWriteData(int len, const u_char* data)
	{
	ASSERT(next_packet);

	while ( len > 0 )
		{
		int left = next_packet->Space();

		if ( ! left )
			{
			PushPacket();
			left = next_packet->Space();
			}

		if ( left > len )
			left = len;

		int res = next_packet->AppendData(data, left);
		ASSERT(res);
		data += left;
		len -= left;
		please_flush = 1;
		}
	}

int TCP_RewriterEndpoint::IsPlaceHolderPacket(TCP_TracePacket* p)
	{
	return p->SeqLength() == 0 &&
	       ! p->GetTCPFlag(TH_SYN | TH_RST | TH_FIN | TH_URG | TH_PUSH) &&
	       ! (p->GetTCPFlag(TH_ACK) && p->GetAck() > last_ack);
	}

void TCP_RewriterEndpoint::PushPacket()
	{
	if ( ! next_packet )
		{
		internal_error(MSG_PREFIX "cannot push packet before packet arrival");
		return;
		}

	// Set sequence number ...
	next_packet->SetSeq(next_seq);
	int seq_len = next_packet->SeqLength();

	next_seq += seq_len;

	// ... and acknowledge peer's recently dumped packet.
	if ( next_packet->GetTCPFlag(TH_SYN | TH_RST) &&
	     ! next_packet->GetTCPFlag(TH_ACK) )
		{
		next_packet->SetTCPFlag(TH_ACK, 0);
		next_packet->SetAck(0);
		}
	else
		{
		next_packet->SetTCPFlag(TH_ACK, 1);
		if ( peer->HasPacket() )
			next_packet->SetAck(peer->NextSeq());
		}

	int RST = next_packet->GetTCPFlag(TH_RST);

	// Enforce the order of timestamps; but delayed FIN is OK
	// when there is a gap.
	if ( next_packet->TS() < network_time &&
	     ! (next_packet->GetTCPFlag(TH_FIN) && there_is_a_gap) )
		{
		Weird(MSG_PREFIX "delayed packet");
		Weird(fmt(MSG_PREFIX "packet time %.6f, dumping time %.6f\n",
			 next_packet->TS(), network_time));
		}

	if ( ! IsPlaceHolderPacket(next_packet) ||
	     ! omit_rewrite_place_holder )
		rewriter->DumpPacket(this, next_packet);

	if ( next_packet->GetTCPFlag(TH_ACK) && next_packet->GetAck() > last_ack )
		last_ack = next_packet->GetAck();

	// Reuse the packet headers.
	next_packet->Reuse();
	please_flush = 0;

	if ( RST )
		{
		Reset(1);	// itself ...
		peer->Reset(0); // ... and peer
		return;
		}

	// Question: do we need to request an ACK?
	if ( seq_len > 0 && network_time > last_packet_time )
		peer->ReqAck();
	}

void TCP_RewriterEndpoint::Weird(const char* name) const
	{
	// rewriter->Weird(name);
	}

TCP_Rewriter::TCP_Rewriter(TCP_Connection* arg_conn, PacketDumper* arg_dumper,
				int arg_MTU, int arg_wait_for_commitment)
	{
	conn = arg_conn;
	dumper = arg_dumper;
	MTU = arg_MTU;
	wait_for_commitment = arg_wait_for_commitment;
	discard_packets = 0;	// till AbortPackets(1);

	orig = new TCP_RewriterEndpoint(this);
	resp = new TCP_RewriterEndpoint(this);

	orig->Init();
	resp->Init();

	anon_addr[0] = anon_addr[1] = 0;

	if ( anonymize_ip_addr )
		{
		anon_addr[0] =
			ip_anonymizer->Anonymize(to_v4_addr(conn->OrigAddr()));
		anon_addr[1] =
			ip_anonymizer->Anonymize(to_v4_addr(conn->RespAddr()));
		}
	}

TCP_Rewriter::~TCP_Rewriter()
	{
	// We don't need to call EndOfData
	// orig->EndOfData();
	// resp->EndOfData();

	if ( ! packet_queue.empty() )
		{
		warn(fmt(MSG_PREFIX
			 "rewriter gets neither commit or abort, "
			 "and %d packets will be discarded", packet_queue.size()));
		AbortPackets(0);
		}

	delete orig;
	delete resp;
	}

void TCP_Rewriter::NextPacket(int is_orig, double t,
			      const struct pcap_pkthdr* pcap_hdr,
			      const u_char* pcap_pkt, int hdr_size,
			      const struct ip* ip,
			      const struct tcphdr* tp)
	{
	TCP_TracePacket* p = new TCP_TracePacket(t, pcap_hdr, MTU);

	int ip_hdr_len = ip->ip_hl * 4;
	int tcp_hdr_len = tp->th_off * 4;

	if ( p->AppendLinkHeader(pcap_pkt, hdr_size) &&
	     p->AppendIPHeader((const u_char*)ip, ip_hdr_len) &&
	     p->AppendTCPHeader((const u_char*)tp, tcp_hdr_len) )
		// Hand the packet to the endpoint.
		Endp(is_orig)->NextPacket(p);
	else
		internal_error(MSG_PREFIX "cannot append headers -- check MTU");
	}

void TCP_Rewriter::WriteData(int is_orig, int len, const u_char* data)
	{
	Endp(is_orig)->WriteData(len, data);
	}

void TCP_Rewriter::WriteData(int is_orig, BroString* str)
	{
	WriteData(is_orig, str->Len(), str->Bytes());
	}

void TCP_Rewriter::SkipGap(int is_orig, int len)
	{
	Endp(is_orig)->SkipGap(len);
	}

void TCP_Rewriter::Push(int is_orig)
	{
	Endp(is_orig)->Push();
	}

void TCP_Rewriter::DumpPacket(TCP_RewriterEndpoint* endp, TCP_TracePacket* p)
	{
	struct pcap_pkthdr* hdr;
	const u_char* pkt;
	int length;
	ipaddr32_t anon_src, anon_dst;		// anonymized IP addresses

	if ( discard_packets )
		return;

	if ( endp == orig )
		{
		anon_src = anon_addr[0];
		anon_dst = anon_addr[1];
		}
	else
		{
		anon_src = anon_addr[1];
		anon_dst = anon_addr[0];
		}

	if ( p->Finish(hdr, pkt, length, anon_src, anon_dst) )
		{
		if ( ! wait_for_commitment )
			dumper->DumpPacket(hdr, pkt, length);
		else
			{
			char* b = new char[sizeof(struct pcap_pkthdr) + length];
			packet_queue.push(b);

			memcpy(b, hdr, sizeof(struct pcap_pkthdr));

			b += sizeof(struct pcap_pkthdr);
			memcpy(b, pkt, length);
			}
		}
	else
		internal_error(MSG_PREFIX "ill formed packet for dumping");
	}

void TCP_Rewriter::ScheduleFIN(int is_orig)
	{
	Endp(is_orig)->ScheduleFIN();
	}

void TCP_Rewriter::AbortPackets(int apply_to_future)
	{
	while ( ! packet_queue.empty() )
		{
		char* p = packet_queue.front();
		packet_queue.pop();
		delete [] p;
		}

	if ( apply_to_future )
		discard_packets = 1;
	}

void TCP_Rewriter::CommitPackets(int apply_to_future)
	{
	while ( ! packet_queue.empty() )
		{
		struct pcap_pkthdr* hdr = (struct pcap_pkthdr*) packet_queue.front();
		packet_queue.pop();

		dumper->DumpPacket(hdr, ((u_char*)hdr) + sizeof(struct pcap_pkthdr), (int)(hdr->caplen));

		delete [] (char*) hdr;
		}

	if ( apply_to_future )
		{ // dump all future packets immediately
		wait_for_commitment = 0;
		discard_packets = 0;
		}
	}

TCP_Endpoint* TCP_Rewriter::GetEndpoint(TCP_RewriterEndpoint* endp)
	{
	if ( endp == orig )
		return conn->Orig();
	else
		return conn->Resp();
	}

TCP_RewriterEndpoint* TCP_Rewriter::GetPeer(TCP_RewriterEndpoint* endp)
	{
	if ( endp == orig )
		return resp;

	else if ( endp == resp )
		return orig;

	else
		return 0;
	}


static queue<TCP_RewriterEndpoint*> rewriters_to_flush;

void schedule_flush(TCP_RewriterEndpoint* endp)
	{
	Ref(endp->Conn());
	rewriters_to_flush.push(endp);
	}

void flush_rewriter_packet()
	{
	while ( ! rewriters_to_flush.empty() )
		{
		TCP_RewriterEndpoint* endp = rewriters_to_flush.front();
		rewriters_to_flush.pop();

		if ( endp )
			endp->Flush();
		Unref(endp->Conn());
		}

	if ( mgr.HasEvents() )
		internal_error("flushing packets generates additional events!");
	}

TCP_Rewriter* get_trace_rewriter(Val* conn_val)
	{
	Connection* conn = (Connection*) conn_val->AsRecordVal()->GetOrigin();
	if ( ! conn || conn->ConnType() != CONN_TCP )
		return 0;

	return ((TCP_Connection*) conn)->TraceRewriter();
	}

#include "common-rw.bif.func_def"
