// $Id: TCP_Rewriter.h,v 1.15 2002/10/26 04:15:44 vern Exp $

#ifndef tcp_rewriter_h
#define tcp_rewriter_h

#include <queue>
using namespace std;

#include <pcap.h>

#include "TCP.h"
#include "Anon.h"

class PacketDumper {
public:
	PacketDumper(pcap_dumper_t* pkt_dump);
	~PacketDumper();

	void DumpPacket(const struct pcap_pkthdr* hdr, const u_char* pkt, int len);

protected:
	pcap_dumper_t* pkt_dump;
	struct timeval last_ts;

	void SortTimeStamp(struct timeval* ts);
};


class TCP_TracePacket {
public:
	TCP_TracePacket(double t, const struct pcap_pkthdr* hdr, int MTU);
	~TCP_TracePacket();

	int AppendLinkHeader(const u_char* chunk, int len);
	int AppendIPHeader(const u_char* chunk, int len);
	int AppendTCPHeader(const u_char* chunk, int len);
	int AppendData(const u_char* chunk, int len);

	// Finish is called before dumping the packet. It sets length
	// fields and computes checksums in TCP/IP headers.
	int Finish(struct pcap_pkthdr*& hdr, const u_char*& pkt, int& length, 
		   ipaddr32_t anon_src, ipaddr32_t anon_dst);
	void Reuse();

	const double TS() const	{ return ts; };	// ?

	const struct pcap_pkthdr* Header() const	{ return &pcap_hdr; };

	const u_char* Buffer() const	{ return pkt; };
	const int Length() const	{ return buffer_offset; };
	const int Space() const		{ return buffer_size - buffer_offset; };
	const int IsReuse() const	{ return reuse; };
	const int Empty() const
		{ return SeqLength() == 0 && ! GetTCPFlag(TH_RST); }

	uint32 GetSeq() const;
	void SetSeq(uint32 seq);

	uint32 GetAck() const;
	void SetAck(uint32 ack);

	int GetTCPFlag(int which) const;
	void SetTCPFlag(int which, int value);

	int SeqLength() const;
	int PayloadLength() const;

protected:
	double ts;
	struct pcap_pkthdr pcap_hdr;
	u_char* pkt;	// of length MTU
	int ip_offset, tcp_offset, data_offset;
	int buffer_size;
	int buffer_offset;
	int reuse;	// whether it is an artificially replicated packet

	void Append(const u_char* chunk, int len);
};

// How a unidirectional flow ends
#define	END_BY_FIN		1
#define	END_BY_RST		2
#define	END_BY_PEER_RST		4

class TCP_Rewriter;
class TCP_RewriterEndpoint {
public:
	TCP_RewriterEndpoint(TCP_Rewriter* rewriter);
	~TCP_RewriterEndpoint();

	void Init();

	// A packet that contains a TCP segment.
	void NextPacket(TCP_TracePacket *p);
	void Deliver(double t, int seq, int len);
	void WriteData(int len, const u_char* data);
	void SkipGap(int len);

	void Push();
	void ReqAck();
	void Flush();
	void EndOfData();
	void ScheduleFIN();
	void Reset(int self);

	uint32 NextSeq() const	{ return next_seq; };
	int HasPacket() const	{ return next_packet != 0; };
	inline Connection* Conn() const;

protected:
	TCP_Rewriter* rewriter;		// TCP rewriter for the connection
	TCP_RewriterEndpoint* peer;	// the peer TCP rewriter endpoint
	TCP_Endpoint* endp;		// the corresponding TCP endpoint

	TCP_TracePacket* next_packet;
	queue<BroString*> prolog;

	double last_packet_time;
	uint32 start_seq;	// start seq -- sent in SYN
	uint32 next_seq;	// seq of next packet
	uint32 last_ack;	// last acknowledgement seq
	int please_flush;
	int flush_scheduled;
	int established;
	int end_of_data;
	int FIN_scheduled;
	int there_is_a_gap;

	// Move onto the next packet header.
	void SetNextPacket(TCP_TracePacket* p);

	void PurgeProlog();

	// Pour data into the current packet (next_packet).
	void DoWriteData(int len, const u_char* data);

	// Push the current packet (next_packet) to dumper.
	void PushPacket();

	// Please flush this endpoint after draining events.
	void ScheduleFlush();

	void GenerateFIN();

	// Whether the packet is a "place holder" packet, i.e. it's
	// harmless to omit the packet (except missing the timestamp
	// it contains)
	int IsPlaceHolderPacket(TCP_TracePacket* p);

	void Weird(const char* name) const;
};

class TCP_Rewriter {
public:
	TCP_Rewriter(TCP_Connection* conn, PacketDumper* dumper, int MTU,
			int wait_for_commitment = 0);
	~TCP_Rewriter();

	// A TCP/IP packet.
	void NextPacket(int is_orig, double t,
			const struct pcap_pkthdr* pcap_hdr,
			const u_char* pcap_pkt,	// link level header
			int hdr_size,		// link level header size
			const struct ip* ip,
			const struct tcphdr* tp);

	void WriteData(int is_orig, int len, const u_char* data);
	void WriteData(int is_orig, const char* data)
		{ WriteData(is_orig, strlen(data), data); }
	void WriteData(int is_orig, int len, const char* data)
		{ WriteData(is_orig, len, (const u_char*) data); }
	void WriteData(int is_orig, BroString* str);
	void SkipGap(int is_orig, int len);

	void Push(int is_orig);
	void ScheduleFIN(int is_orig);

	void DumpPacket(TCP_RewriterEndpoint* endp, TCP_TracePacket* p);

	// When wait_for_commitment = 1, packets are not dumped until
	//   CommitPackets().
	// When apply_to_future = 1, the same decision holds for future
	//   packets as well.
	//
	// Regarding why AbortPackets() takes an apply_to_future flag:
	//
	//	The model is that there can be multiple commit/abort stages
	//	during the course of a connection.  At the end of each
	//	stage, a commit or abort decision is made for packets
	//	generated during the stage.  A possible scenario is that
	//	user may want to delete a middle part of a conversation
	//	while keeping the parts before and after intact, and cannot
	//	make the decision until the end of the middle part.

	void AbortPackets(int apply_to_future);
	void CommitPackets(int apply_to_future);

	TCP_Connection* Conn() const	{ return conn; }

	TCP_Endpoint* GetEndpoint(TCP_RewriterEndpoint* endp);
	TCP_RewriterEndpoint* GetPeer(TCP_RewriterEndpoint* endp);

	void Weird(const char* name) const
		{ conn->Weird(name); }

protected:
	TCP_RewriterEndpoint* Endp(int is_orig) const
		{ return is_orig ? orig : resp; }

	TCP_Connection* conn;
	PacketDumper* dumper;
	TCP_RewriterEndpoint* orig;
	TCP_RewriterEndpoint* resp;
	int MTU;
	int wait_for_commitment;
	int discard_packets;
	queue<char*> packet_queue;
	ipaddr32_t anon_addr[2];
};


inline Connection* TCP_RewriterEndpoint::Conn() const
	{
	return rewriter->Conn();
	};

// Please flush the rewriter endpoint after event processing.
void schedule_flush(TCP_RewriterEndpoint* endp);

void flush_rewriter_packet();

extern PacketDumper* transformed_pkt_dump;
extern int transformed_pkt_dump_MTU;

extern TCP_Rewriter* get_trace_rewriter(Val* conn_val);

#if 0 /* draft */
enum tcp_data_encoding {
	ENCODING_PLAIN,
	ENCODING_INT,
	ENCODING_QUOTED_STRING,
};

class TCP_Chunk {
public:
	TCP_Chunk(int begin_seq, int end_seq, data_chunk_t net_data);
	~TCP_Chunk();

	// int Decode(data_type& data);
	// int Encode(data_type n);

protected:
	// data location in the original trace
	int begin_seq, end_seq;

	BroString* data;
	int encoding;
	void* aux;		// for additional info
};
#endif /* draft */

#endif
