# $Id: ftp.bro,v 1.108 2002/10/20 02:35:20 vern Exp $

@load alert
@load conn
@load scan
@load hot-ids

redef enum Alert += {
	FTP_UnexpectedConn,	# FTP data transfer from unexpected source
	FTP_ExcessiveFilename,	# very long filename seen
	FTP_PrivPort,		# privileged port used in PORT/PASV; $sub says which
	FTP_BadPort,		# bad format in PORT/PASV; $sub says which
	FTP_Sensitive,		# sensitive connection - not more specific
};

redef capture_filter += "port ftp";

type ftp_session_info: record {
	id: count;
	connection_id: conn_id;
	user: string;
	request: string;	# pending request or requests
	num_requests: count;	# count of pending requests
	request_t: time;	# time of request
	log_if_not_denied: bool;	# unless code 530 on reply, log it
	log_if_not_unavail: bool;	# unless code 550 on reply, log it
	log_it: bool;		# if true, log the request(s)
};

global ftp_log = open_log_file("ftp") &redef;

global ftp_sessions: table[conn_id] of ftp_session_info;
global ftp_session_id = 0;

# Indexed by the responder pair, yielding the address expected to connect to it.
global ftp_data_expected: table[addr, port] of addr;

# The associated session ID number.
global ftp_data_expected_session: table[addr, port] of ftp_session_info;

const ftp_guest_ids = { "anonymous", "ftp", "guest", } &redef;

# Indexed by source & destination addresses and the id.
const ftp_skip_hot: set[addr, addr, string] &redef;

const ftp_hot_files =
	  /.*(etc\/|master\.)?(passwd|shadow|s?pwd\.db)/
	&redef;

const ftp_hot_guest_files =
	  /.*\.rhosts/
	| /.*\.forward/
	&redef;

const ftp_hot_cmds: table[string] of pattern = {
	["SITE"] = /[Ee][Xx][Ee][Cc].*/,
} &redef;

global ftp_excessive_filename_len = 250 &redef;
global ftp_excessive_filename_trunc_len = 32 &redef;

# Invalid PORT/PASV directives that exactly match the following
# don't generate alerts.
global ftp_ignore_invalid_PORT =
	/,0,0/	# these are common, dunno why
&redef;

# Some servers generate particular privileged PASV ports for benign
# reasons (presumably to tunnel through firewalls, sigh).
global ftp_ignore_privileged_PASVs = { ssh, } &redef;

const ftp_file_cmds = {
	"APPE", "CWD", "DELE", "MKD", "RETR", "RMD", "RNFR", "RNTO",
	"STOR", "STOU",
};

const ftp_skip_replies = {
	150,	# "status okay - about to open connection"
	331	# "user name okay, need password"
};

const ftp_replies: table[count] of string = {
	[150] = "ok",
	[200] = "ok",
	[220] = "ready for new user",
	[221] = "closed",
	[226] = "complete",
	[230] = "logged in",
	[250] = "ok",
	[257] = "done",
	[331] = "id ok",
	[500] = "syntax error",
	[530] = "denied",
	[550] = "unavail"
};

# Pairs of IP addresses for which we shouldn't bother logging if one
# of them is used in lieu of the other in a PORT or PASV directive.

const skip_unexpected: set[addr] = {
	15.253.0.10, 15.253.48.10, 15.254.56.2,		# hp.com
	gvaona1.cns.hp.com,
} &redef;

const skip_unexpected_net: set[addr] &redef;

event ftp_unexpected_conn_violation(c: connection, orig: addr, expected: addr)
	{
	ALERT([$alert=FTP_UnexpectedConn, $conn=c,
		$msg=fmt("%s > %s (%s) FTP connection from %s",
			c$id$orig_h, c$id$resp_h, c$addl, orig)]);
	}

event ftp_unexpected_conn(c: connection, orig: addr, expected: addr)
	{
	if ( orig in skip_unexpected || expected in skip_unexpected ||
	     mask_addr(orig, 24) in skip_unexpected_net ||
	     mask_addr(expected, 24) in skip_unexpected_net )
		; # don't bother reporting

	else if ( mask_addr(orig, 24) == mask_addr(expected, 24) )
		; # close enough, probably multi-homed

	else if ( mask_addr(orig, 16) == mask_addr(expected, 16) )
		; # ditto

	else
		event ftp_unexpected_conn_violation(c, orig, expected);
	}

function is_ftp_data_conn(c: connection): bool
	{
	local id = c$id;
	if ( [id$resp_h, id$resp_p] in ftp_data_expected )
		{
		local expected = ftp_data_expected[id$resp_h, id$resp_p];
		if ( id$orig_h != expected )
			{
			local session = ftp_data_expected_session[id$resp_h, id$resp_p];
			local session_conn = lookup_connection(session$connection_id);
			event ftp_unexpected_conn(session_conn, id$orig_h, expected);
			}

		delete ftp_data_expected[id$resp_h, id$resp_p];
		delete ftp_data_expected_session[id$resp_h, id$resp_p];
		return T;
		}

	else if ( id$orig_p == 20/tcp &&
	          [id$resp_h, id$resp_p, id$orig_h, 21/tcp] in ftp_sessions )
		return T;
	else
		return F;
	}

function new_ftp_session(c: connection): any
	{
	local session = c$id;
	local new_id = ++ftp_session_id;

	local info: ftp_session_info;
	info$id = new_id;
	info$connection_id = session;
	info$user = "<unknown>";
	info$request = "";
	info$num_requests = 0;
	info$request_t = c$start_time;
	info$log_if_not_unavail = F;
	info$log_if_not_denied = F;
	info$log_it = F;

  	ftp_sessions[session] = info;
  	c$addl = fmt("#%d", new_id);

	print ftp_log, fmt("%.6f #%d %s start", c$start_time, new_id,
				id_string(session));
	}

function ftp_message(id: conn_id, msg: string): bool
	{
	print ftp_log, fmt("%.6f #%d %s",
			network_time(), ftp_sessions[id]$id, msg);
	}

event ftp_sensitive_file(session: ftp_session_info, filename: string)
	{
	session$log_if_not_unavail = T;
	}

event ftp_excessive_filename(session: ftp_session_info,
				command: string, arg: string)
	{
	ALERT([$alert=FTP_ExcessiveFilename, $id=session$connection_id,
		$msg=fmt("%s #%d excessive filename: %s",
			id_string(session$connection_id), session$id, arg)]);
	session$log_it = T;
	}

event ftp_request(c: connection, command: string, arg: string)
	{
	local id = c$id;
	# cmd and arg for transformed trace
	local rw_cmd = command;
	local rw_arg = arg;

	if ( id !in ftp_sessions )
		new_ftp_session(c);

	local session = ftp_sessions[id];

	if ( command == "USER" )
		{
		if ( arg in hot_ids &&
		     [id$orig_h, id$resp_h, arg] !in ftp_skip_hot )
			{
			if ( arg in always_hot_ids )
				session$log_it = T;
			else
				session$log_if_not_denied = T;
			}

		append_addl(c, arg);
		session$user = arg;

		if ( arg in forbidden_ids )
			terminate_connection(c);
		}

	else if ( command == "PASS" )
		{
		if ( session$user in forbidden_ids_if_no_password &&
		     arg == "" )
			terminate_connection(c);

		if ( session$user in ftp_guest_ids )
			append_addl_marker(c, arg, "/");
		else
			{
			event account_tried(c, session$user, arg);
			arg = "<suppressed>";
			}

		rw_arg = "*** password ***";
		}

	else if ( command == "PORT" )
		{
		local data = parse_ftp_port(arg);
		if ( data$valid )
			{
			if ( data$h != id$orig_h )
				ftp_message(id, fmt("*> PORT host %s doesn't match originator host %s", data$h, id$orig_h));

			if ( data$p < 1024/tcp && data$p in port_names )
				ALERT([$alert=FTP_PrivPort, $id=id,
					$msg=fmt("%s #%d privileged PORT %d: %s",
						id_string(id), session$id,
						data$p, arg),
					$sub="PORT"]);

			ftp_data_expected[data$h, data$p] = c$id$resp_h;
			ftp_data_expected_session[data$h, data$p] = session;
			}
		else if ( arg != ftp_ignore_invalid_PORT )
			ALERT([$alert=FTP_BadPort, $id=id,
				$msg=fmt("%s #%d invalid ftp PORT directive: %s",
					id_string(id), session$id, arg),
				$sub="PORT"]);
		}

	else if ( command in ftp_file_cmds )
		{
		if ( (arg == ftp_hot_files ||
		      (session$user in ftp_guest_ids &&
		       arg == ftp_hot_guest_files)) )
			event ftp_sensitive_file(session, arg);

		if ( byte_len(arg) >= ftp_excessive_filename_len )
			{
			arg = fmt("%s..[%d]..",
				sub_bytes(arg, 1, ftp_excessive_filename_trunc_len),
				byte_len(arg));
			event ftp_excessive_filename(session, command, arg);
			}
		}

	else if ( command == "ACCT" )
		append_addl(c, fmt("(account %s)", arg));

	if ( command in ftp_hot_cmds && arg == ftp_hot_cmds[command] )
		{
		session$log_it = T;

		# Special hack for "site exec" attacks.
		### Obviously, this should be generic and not specialized
		### like the following.
		if ( command == "SITE" && /[Ee][Xx][Ee][Cc]/ in arg &&
		     # We see legit use of "site exec cp / /", God knows why.
		     byte_len(arg) > 32 )
			{ # Terminate with extreme prejudice.
			terminate_connection(c);
			drop_address(c$id$orig_h, fmt("%s %s", command, arg));
			}
		}

	local request = arg == "" ? command : cat(command, " ", arg);
	if ( ++session$num_requests == 1 )
		{
		# First pending request
		session$request = request;
		session$request_t = network_time();
		}
	else
		{
		# Don't append PASS commands, unless they're for an
		# anonymous user.
		if ( command == "PASS" )
			{
			if ( session$user in ftp_guest_ids )
				{
				session$request =
					cat(session$request, "/", arg);
				}

			# Don't count this as a multiple request.
			--session$num_requests;
			}
		else
			{
			if ( byte_len(session$request) < 256 )
				session$request = cat(session$request, ", ", request);
			}
		}

	### Is it okay to include the args of an ACCT command?
	rewrite_ftp_request(c, rw_cmd, rw_arg);
	}

event ftp_binary_response(session: ftp_session_info, code: count, msg: string)
	{
	print ftp_log, fmt("%.6f #%d binary response",
				network_time(), session$id);
	}

event ftp_reply(c: connection, code: count, msg: string, cont_resp: bool)
	{
	rewrite_ftp_reply(c, code, msg, cont_resp);

	if ( cont_resp )
		return;

	local id = c$id;

	if ( id !in ftp_sessions )
		new_ftp_session(c);

	local session = ftp_sessions[id];

	if ( session$log_if_not_denied && code != 530 &&
	     # skip password prompt, which we can get when the requests
	     # are stacked up
	     code != 331 )
		session$log_it = T;

	if ( session$log_if_not_unavail && code != 550 )
		session$log_it = T;

	if ( code == 227 )
		{
		local data = parse_ftp_pasv(msg);
		if ( data$valid )
			{
			if ( data$h != id$resp_h )
				ftp_message(id, fmt("*< PASV host %s doesn't match responder host %s", data$h, id$resp_h));

			if ( data$p < 1024/tcp && data$p in port_names &&
			     data$p !in ftp_ignore_privileged_PASVs )
				ALERT([$alert=FTP_PrivPort, $id=id,
					$msg=fmt("%s #%d privileged PASV %d: %s",
						id_string(id), session$id,
						data$p, msg),
					$sub="PASV"]);

			ftp_data_expected[data$h, data$p] = c$id$orig_h;
			ftp_data_expected_session[data$h, data$p] = session;

			msg = endpoint_id(data$h, data$p);
			}
		else if ( msg != ftp_ignore_invalid_PORT )
			{
			ALERT([$alert=FTP_BadPort, $id=id,
				$msg=fmt("%s #%d invalid ftp PASV directive: %s",
					id_string(id), session$id, msg),
				$sub="PASV"]);
			msg = "invalid PASV";
			}
		}

	if ( session$num_requests > 0 )
		{
		if ( code in ftp_skip_replies )
			;	# Don't flush request yet.

		else
			{
			local reply = code in ftp_replies ? ftp_replies[code] :
					fmt("%d %s", code, msg);

			local session_msg = fmt("#%d %s%s (%s)",
					session$id,
					session$num_requests > 1 ? "*" : "",
					session$request, reply);

			if ( session$log_it )
				ALERT([$alert=FTP_Sensitive, $id=id,
					$msg=fmt("ftp: %s %s",
						id_string(id), session_msg)]);

			print ftp_log, fmt("%.6f %s", session$request_t,
						session_msg);

			session$request = "";
			session$num_requests = 0;
			session$log_if_not_unavail = F;
			session$log_if_not_denied = F;
			session$log_it = F;
			}
		}
	else
		{
		# An unpaired response.  This can happen in particular
		# when the session is encrypted, so we check for that here.
		if ( /[\x80-\xff]{3}/ in msg )
			# Three 8-bit characters in a row - good enough.
			# Note, this should of course be customizable.
			event ftp_binary_response(session, code, msg);

		else
			print ftp_log, fmt("%.6f #%d response (%d %s)",
					network_time(), session$id, code, msg);
		}
	}

event connection_finished(c: connection)
	{
	local id = c$id;
	if ( id in ftp_sessions )
		{
		local session = ftp_sessions[id];

		if ( session$num_requests > 0 )
			{
			local msg = fmt("#%d %s%s (no reply)", session$id,
				session$num_requests > 1 ? "*" : "",
				session$request);

			if ( session$log_it )
				ALERT([$alert=FTP_Sensitive, $id=id,
					$msg=fmt("ftp: %s %s",
						id_string(id), msg)]);

			print ftp_log, fmt("%.6f %s", session$request_t, msg);
			}

		ftp_message(id, "finish");
		delete ftp_sessions[id];
		}
	}

event bro_init()
	{
	have_FTP = T;
	}

function is_ftp_conn(c: connection): bool
	{
	return c$id$resp_p == ftp;
	}

event connection_state_remove(c: connection)
	{
	if ( requires_trace_commitment && is_ftp_conn(c) )
		# commit queued packets and all packets in future
		rewrite_commit_trace(c, T, T);
	}
