# $Id: rules.bro,v 1.11 2002/10/20 02:35:20 vern Exp $

@load alert

redef enum Alert += {
	SensitiveRule,		# generic for triggering a log-worthy rule
	MultipleRules,		# host has triggered many signatures
};

const RULE_IGNORE = 0;		# ignore this rule completely (even for scan detection)
const RULE_QUIET = 1;		# process but don't report individually
const RULE_FILE = 2;		# write into the rule.log file
const RULE_LOG = 3;		# log and write into the rule.log file

# Action for a rule.
const rule_actions: table[string] of count =  {

	["sid-526"] = RULE_IGNORE,
		# sid-526 BAD TRAFFIC data in TCP SYN packet
	
	["sid-615"] = RULE_QUIET,
		# sid-615 SCAN SOCKS Proxy attempt

	["sid-618"] = RULE_QUIET,
		# sid-618 SCAN Squid Proxy attempt

	["sid-620"] = RULE_QUIET,
		# sid-620 SCAN Proxy (8080) attempt

	["sid-823"] = RULE_QUIET,
		# sid-823 WEB-CGI cvsweb.cgi access
        
	["sid-872"] = RULE_QUIET,
		# sid-872 WEB-CGI tcsh access
	
	["sid-1002"] = RULE_QUIET,
		# sid-1002 WEB-IIS cmd.exe access

	["sid-1113"] = RULE_QUIET,
		# sid-1113 WEB-MISC http directory traversal
	
	["sid-1214"] = RULE_QUIET,
		# sid-1214 WEB-MISC intranet access

	["sid-1287"] = RULE_QUIET,
		# sid-1287 WEB-IIS scripts access
	
	["sid-1288"] = RULE_QUIET,
		# sid-1288 WEB-FRONTPAGE /_vti_bin/ access

	["sid-1560"] = RULE_QUIET,
		# sid-1560 WEB-MISC /doc/ access

	["sid-1653"] = RULE_QUIET,
		# sid-1653 WEB-CGI campus access

} &redef &default = RULE_FILE;

global rule_file = open_log_file("rules");

# Scan detection.

# Log if for a pair (orig, rule) the number of different responders has
# reached one of the thresholds.
const horiz_scan_thresholds : set[count] = { 5, 10, 50, 100, 500, 1000 } &redef;

# Log if for a pair (orig, resp) the number of different rule matches has
# reached one of the thresholds.
const vert_scan_thresholds : set[count] = { 5, 10, 50, 100, 500, 1000 } &redef;

type rule_set: set[string];
type addr_set: set[addr];

# You may want to define some &read_expires on these:
global horiz_table: table[addr, string] of addr_set;
global vert_table: table[addr, addr] of rule_set;
global last_hthresh: table[addr] of count &default = 0;
global last_vthresh: table[addr] of count &default = 0;

function log_scan_and_file(txt: string)
	{
	local msg = fmt("%f %s", network_time(), txt);
	ALERT([$alert=MultipleRules, $msg=msg]);
	print rule_file, msg;
	}

event rule_matched(state: rule_state, msg: string, data: string)
	{
	local id = state$id;
	local action = rule_actions[id];

	if ( action == RULE_IGNORE )
		return;

	# We always add it to the connection record.
	append_addl(state$conn, state$id);

	if ( action == RULE_FILE || action == RULE_LOG )
		print rule_file, fmt("%f %s/%d > %s/%d %s %s", network_time(),
				state$conn$id$orig_h, state$conn$id$orig_p,
				state$conn$id$resp_h, state$conn$id$resp_p,
				state$id, msg);

	if ( action == RULE_LOG )
		ALERT([$alert=SensitiveRule, $conn=state$conn,
			$msg=fmt("%s > %s rule match for %s: %s",
				state$conn$id$orig_h, state$conn$id$resp_h,
				state$id, msg)]);

	# Keep track of scans
	local orig = state$conn$id$orig_h;
	local resp = state$conn$id$resp_h;

	if ( [orig, id] !in horiz_table )
		{
		local emptyset1: addr_set;
		horiz_table[orig, id] = emptyset1;
		}
	add horiz_table[orig, id][resp];

	if ( [orig, resp] !in vert_table )
		{
		local emptyset2: rule_set;
		vert_table[orig, resp] = emptyset2;
		}
	add vert_table[orig, resp][id];

	local hcount = length(horiz_table[orig, id]);
	local vcount = length(vert_table[orig, resp]);

	if ( hcount in horiz_scan_thresholds && hcount != last_hthresh[orig] )
		{
		log_scan_and_file(fmt("%s has triggered rule %s on %d hosts",
				orig, id, hcount));
		last_hthresh[orig] = hcount;
		}

	if ( vcount in vert_scan_thresholds && vcount != last_vthresh[orig] )
		{
		log_scan_and_file(fmt("%s has triggered %d different rules on host %s",
				orig, vcount, resp));
		last_vthresh[orig] = vcount;
		}
	}

# Returns true if the given rule has already been triggered for the given
# orig/resp pair.
function rule_has_matched(id: string, orig: addr, resp: addr): bool
	{
	return [orig, resp] in vert_table ? id in vert_table[orig, resp] : F;
	}
